"""Tuning tools."""
# flake8: noqa

from .__version__ import *
from ._arrangement import *
from ._discrete_tune import *
from ._holistic import *
from ._instrument import *
from ._intensity import *
from ._map import *
from ._setable import *
from ._note import *
from ._offset import *
from ._open import *
from ._rename import *
from ._setpoint import *
from ._store import *
from ._tune import *
from ._tune_test import *
from ._update_merge import *
from .io import *



__all__ = ["rename"]

from ._transition import Transition, TransitionType
from ._instrument import Instrument


def rename(instr: Instrument, name: str):
    """Rename an instrument.

    Note: this tranistion breaks the history, as the primary key changes.
    """
    trans = Transition(TransitionType.rename, metadata={"old_name": instr.name})
    new_instr = instr.as_dict()
    new_instr["transition"] = trans
    new_instr["name"] = name
    return Instrument(**new_instr)



__all__ = ["Instrument"]


from datetime import datetime as _datetime
from typing import Dict, Optional, Union
import json

from ._arrangement import Arrangement
from ._setable import Setable
from ._note import Note
from ._transition import Transition, TransitionType


class Instrument(object):
    def __init__(
        self,
        arrangements: Dict[str, Union[Arrangement, dict]],
        setables: Dict[str, Optional[Union[Setable, dict]]] = None,
        *,
        name: Optional[str] = None,
        transition: Optional[Union[Transition, dict]] = None,
        load: Optional[float] = None,
    ):
        """Representation of a system of arrangements for an instrument.

        Parameters
        ----------
        arrangements: Dict[str, Union[Arrangement, dict]
            Dictionary of arrangements in the instrument
        setables: Dict[str, Optional[Union[Setable, dict]]]
            Default values for the instrument. Can be ignored unless you
            require your instrument to have default positions.
        name: Optional[str]
            The name of the instrument, used to store/retrieve the instrument.
        transition: Optional[Union[Transition, dict]]
            The operation which creates this instrument.
            transitions are concise records for instrument changes.
            If not given, a blank Transition (TransitionType "create") will be
            made.
        load: Optional[float]
            POSIX timestamp of the tune when retrieved from the store.
            Ignore for instruments not retrieved from the store.
        """
        self._name: Optional[str] = name
        self._arrangements: Dict["str", Arrangement] = {
            k: Arrangement(**v) if isinstance(v, dict) else v for k, v in arrangements.items()
        }
        if setables is None:
            setables = {}
        self._setables: Dict["str", Setable] = {
            k: Setable(**v) if isinstance(v, dict) else v for k, v in setables.items()
        }
        if transition is None:
            self._transition = Transition(TransitionType.create)
        elif isinstance(transition, dict):
            self._transition = Transition(**transition)
        else:
            self._transition = transition
        self._load: Optional[float] = load

    def __repr__(self):
        ret = f"Instrument({repr(self.arrangements)}, {repr(self.setables)}"
        if self.name is not None:
            ret += f", name={repr(self.name)}"
        if self.transition.type != TransitionType.create:
            ret += f", transition={repr(self.transition)}"
        return ret + ")"

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self._setables != other._setables:
            return False
        if self._arrangements != other._arrangements:
            return False
        return True

    def __call__(self, ind_value, arrangement_name=None) -> Note:
        # get correct arrangement
        valid = []
        for arrangement in self._arrangements.values():
            # we should probably do "close enough" for floating point on the edges...
            try:
                if arrangement.ind_min <= ind_value <= arrangement.ind_max:
                    valid.append(arrangement)
            except ValueError:
                if (arrangement.ind_min <= ind_value).all() and (
                    ind_value <= arrangement.ind_max
                ).all():
                    valid.append(arrangement)
        if arrangement_name is not None:
            assert arrangement_name in [v.name for v in valid]
            arrangement = self._arrangements[arrangement_name]
        elif len(valid) == 1:
            arrangement = valid[0]
        elif len(valid) == 0:
            raise ValueError(f"There are no valid arrangements at {ind_value}.")
        else:
            raise ValueError("There are multiple valid arrangements! You must specify one.")
        # call arrangement
        setable_positions = {}
        setables = self._setables.copy()
        todo = [(ind_value, tune) for tune in arrangement.tunes.items()]
        while todo:
            v, t = todo.pop(0)
            tune_name, tune = t
            if tune_name in self._arrangements:
                new = [
                    (tune(v), subtune) for subtune in self._arrangements[tune_name].tunes.items()
                ]
                todo += new
            else:
                # Since todos are appended for any subarrangement, the top-most (eg idl in idl>sig)
                # Arrangment should be used, so skip setting if it is defined in inner arrangments
                if tune_name in setable_positions:
                    continue
                setable_positions[tune_name] = tune(v)
                setables[tune_name] = Setable(tune_name)
        for setable in self._setables:
            if setable not in setable_positions and self._setables[setable].default is not None:
                setable_positions[setable] = self._setables[setable].default
        # finish
        note = Note(
            setables=self._setables,
            setable_positions=setable_positions,
            arrangement_name=arrangement.name,
        )
        return note

    def __getitem__(self, item):
        return self._arrangements[item]

    def as_dict(self):
        """Dictionary representation for this Instrument."""
        out = {}
        out["name"] = self.name
        out["arrangements"] = {k: v.as_dict() for k, v in self._arrangements.items()}
        out["setables"] = {k: v.as_dict() for k, v in self._setables.items()}
        out["transition"] = self.transition.as_dict()
        return out

    @property
    def name(self):
        """The name of the instrument.

        This is the key that is used to store/retrieve the instrument.
        """
        return self._name

    @property
    def transition(self):
        """The transition operation that generated this instrument."""
        return self._transition

    @property
    def setables(self):
        """The setables associated with this instrument."""
        return self._setables

    @property
    def arrangements(self):
        """The arrangements associated with this instrument."""
        return self._arrangements

    @property
    def load(self):
        """The POSIX timestamp for when this instrument was created, if it was stored."""
        return self._load

    def save(self, file):
        """Save the JSON representation into an open file."""

        class NdarrayEncoder(json.JSONEncoder):
            def default(self, obj):
                if hasattr(obj, "tolist"):
                    return obj.tolist()
                return json.JSONEncoder.default(self, obj)

        json.dump(self.as_dict(), file, cls=NdarrayEncoder)



__all__ = ["map_ind_points", "map_ind_limits"]

import copy

import numpy as np

from ._transition import Transition
from ._tune import Tune


def map_ind_points(instrument, arrangement, tune, setpoints, units=None):
    """Map the independent values of a tune onto new setpoints.

    Parameters
    ----------
    instrument: Instrument
        The instrument object to alter.
    arrangement: str
        The name of the arrangemnt to alter.
    tune: str
        The name of the tune to alter.
    setpoints: array-like
        The new setpoints to map the tune to.
    units: Optional[str]
        The units of the new setpoints, will be converted to the original units.
        If not given, original units are assumed.

    Returns
    -------
    Instrument
        The instrument with the tune remapped to new setpoints.
    """
    md = {"arrangement": arrangement, "tune": tune, "setpoints": setpoints, "units": units}
    to_replace = instrument[arrangement][tune]
    if units is not None:
        setpoints = wt.units.convert(setpoints, units, to_replace.ind_units)
    instr = copy.deepcopy(instrument)
    instr[arrangement]._tunes[tune] = Tune(
        setpoints, to_replace(setpoints), dep_units=to_replace.dep_units
    )
    instr._transition = Transition("map_ind_points", instrument, metadata=md)
    instr._load = None
    return instr


def map_ind_limits(instrument, arrangement, tune, min, max, units=None):
    """Map the independent values of a tune onto new limits.

    The resultant tune will have the same number of points as the original,
    evenly spaced between ``min`` and ``max``.

    Parameters
    ----------
    instrument: Instrument
        The instrument object to alter.
    arrangement: str
        The name of the arrangemnt to alter.
    tune: str
        The name of the tune to alter.
    min: float
        The new minimum setpoint.
    max: float
        The new maximum setpoint.
    units: Optional[str]
        The units of the new setpoints, will be converted to the original units.
        If not given, original units are assumed.

    Returns
    -------
    Instrument
        The instrument with the tune remapped to new setpoints.
    """
    to_replace = instrument[arrangement][tune]
    points = np.linspace(min, max, len(to_replace))
    instr = map_ind_points(instrument, arrangement, tune, points, units)
    md = {"arrangement": arrangement, "tune": tune, "min": min, "max": max, "units": units}
    instr._transition = Transition("map_ind_limits", instrument, metadata=md)
    return instr



__all__ = ["open"]

import json

from ._instrument import Instrument

open_ = open


def open(path, *, load=False):
    """Open an instrument stored in a JSON file.

    Parameters
    ----------

    path: PathLike or FileLike
        The path to a file which contains an instrument
    load: datetime
        Allows this method to be used for loading by providing its associated store time
        Should generally be avoided when used directly

    Returns
    -------
    Instrument
        The instrument that was stored in the file
    """
    if hasattr(path, "read"):
        d = json.load(path)
    else:
        with open_(path, "r") as f:
            d = json.load(f)

    return Instrument(**d, load=load)



import pathlib

import WrightTools as wt


def save(instrument, fig, image_name, save_directory=None):
    if save_directory is None:
        save_directory = "."
    save_directory = pathlib.Path(save_directory)
    with open(save_directory / "instrument.json", "w") as f:
        instrument.save(f)
    # Should we timestamp the image?
    p = (save_directory / image_name).with_suffix(".png")
    wt.artists.savefig(p, fig=fig)



"""Tools to interact with the attune store."""

__all__ = ["catalog", "load", "restore", "store", "undo"]


from datetime import datetime, timedelta, timezone
import pathlib
import os
import warnings

import appdirs
import dateutil

from ._transition import Transition, TransitionType
from ._open import open as open_


def catalog(full=False):
    """Access a catalog of instruments.

    By default returns a list of keys available.
    If full is True, loads each instrument as a dictionary of keys to Instrument objects.
    """
    if "ATTUNE_STORE" in os.environ and os.environ["ATTUNE_STORE"]:
        attune_dir = pathlib.Path(os.environ["ATTUNE_STORE"])
    else:
        attune_dir = pathlib.Path(appdirs.user_data_dir("attune", "attune"))
    instrument_names = os.listdir(attune_dir)
    if full:
        return {name: load(name) for name in instrument_names}
    else:
        return instrument_names


def load(name: str, time=None, reverse: bool = True):
    """Load an istrument of the given name.

    Parameters
    ----------
    name: str
        The key of the instrument to load
    time: str, datetime, optional
        The time for which to load the instrument.
        Allows loading previous instruments in the catalog.
        Allows for some natural language descriptions e.g. "5 minutes ago".
        By default uses the current timestamp.
    reverse: boolean, optional
        Direction to search, by default looks for a previous curve.
        If given as False, looks forward in time from the given timestamp.
    """
    if isinstance(time, str):
        import maya

        time = maya.when(time)
    if time is None:
        time = datetime.now(timezone.utc)
    if hasattr(time, "datetime"):
        time = time.datetime()

    def find(name, time, reverse):
        year = time.year
        month = time.month

        if "ATTUNE_STORE" in os.environ and os.environ["ATTUNE_STORE"]:
            attune_dir = pathlib.Path(os.environ["ATTUNE_STORE"])
        else:
            attune_dir = pathlib.Path(appdirs.user_data_dir("attune", "attune"))

        if not (attune_dir / name).exists():
            raise ValueError(f"No instrument found with name '{name}'")

        while True:
            datadir = attune_dir
            datadir /= name
            datadir /= str(year)
            datadir /= f"{month:02}"
            if datadir.exists():
                for d in sorted(
                    datadir.iterdir(),
                    key=lambda x: dateutil.parser.isoparse(x.name),
                    reverse=reverse,
                ):
                    if reverse:
                        if dateutil.parser.isoparse(d.name) <= time:
                            return datadir / d.name
                    else:
                        if dateutil.parser.isoparse(d.name) >= time:
                            return datadir / d.name

            if reverse:
                if month == 1:
                    year -= 1
                    month = 12
                else:
                    month -= 1
                if year < 1960:
                    raise ValueError(
                        f"Could not find an instrument earlier than {time}. Looked back all the way to the invention of the laser"
                    )
            else:
                if month == 12:
                    month = 1
                    year += 1
                else:
                    month += 1
                if year > datetime.now().year + 20:
                    raise ValueError(f"Could not find an instrument later than {time}.")

    datadir = find(name, time, reverse)
    return open_(datadir / "instrument.json", load=dateutil.parser.isoparse(datadir.name))


def restore(name, time, reverse=True):
    """Restore a previously applied instrument.


    Parameters
    ----------
    name: str
        The key of the instrument to load
    time: str, datetime
        The time for which to load the instrument.
        Allows loading previous instruments in the catalog.
        Allows for some natural language descriptions e.g. "5 minutes ago".
        By default uses the current timestamp.
    reverse: boolean, optional
        Direction to search, by default looks for a previous curve.
        If given as False, looks forward in time from the given timestamp.
    """
    instr = load(name, time, reverse)
    if load(name) == instr:
        warnings.warn("Attempted to restore instrument equivalent to current head, ignoring.")
        return
    instr._transition = Transition(
        TransitionType.restore, metadata={"time": instr.load.isoformat()}
    )
    _store_instr(instr)


def store(instrument, warn=True):
    """Store an instrument into the catalog.

    Parameters
    ----------
    instrument: Instrument
        The instrument to store.
    warn: bool
        Whether or not to warn if the store is equivalent to the current head.
    """
    try:
        if load(instrument.name) == instrument:
            if warn:
                warnings.warn(
                    "Attempted to store instrument equivalent to current head, ignoring."
                )
            return
    except ValueError:
        pass  # Could mean it is not yet in store at all

    if instrument.load is None and instrument.transition.previous is not None:
        store(instrument.transition.previous, warn=False)

    if instrument.load is not None:
        restore(instrument.name, instrument.load)
        return

    _store_instr(instrument)


def _store_instr(instrument):
    if "ATTUNE_STORE" in os.environ and os.environ["ATTUNE_STORE"]:
        attune_dir = pathlib.Path(os.environ["ATTUNE_STORE"])
    else:
        attune_dir = pathlib.Path(appdirs.user_data_dir("attune", "attune"))

    while True:
        now = datetime.now(timezone.utc)
        # make datadir
        datadir = attune_dir
        datadir /= instrument.name
        datadir /= f"{now.year}"
        datadir /= f"{now.month:02}"
        datadir /= now.isoformat(timespec="milliseconds").replace("-", "").replace(":", "")
        try:
            datadir.mkdir(parents=True)
        except FileExistsError:
            continue
        else:
            break
    # store instrument
    with open(datadir / "instrument.json", "w") as f:
        instrument.save(f)
        f.write("\n")
    # store data
    if instrument.transition.data is not None:
        instrument.transition.data.save(datadir / "data.wt5")
    # store old instrument
    if instrument.transition.previous is not None:
        with open(datadir / "previous_instrument.json", "w") as f:
            instrument.transition.previous.save(f)


def undo(instrument):
    """Undo one transition."""
    if instrument.load is not None:
        return load(instrument.name, instrument.load - timedelta(milliseconds=1))
    elif instrument.transition.previous is not None:
        return instrument.transition.previous
    raise ValueError("Nothing to undo")



"""Workup an instrument maximizing intensity of a single setable"""

import numpy as np
import WrightTools as wt

from ._instrument import Instrument
from ._arrangement import Arrangement
from ._setable import Setable
from ._tune import Tune
from ._transition import Transition
from ._plot import plot_intensity
from ._common import save


# --- processing methods --------------------------------------------------------------------------

__all__ = ["intensity"]


def _intensity(data, channel_name, tune_points, *, spline=True, **spline_kwargs):
    data.moment(axis=1, channel=channel_name, moment=1, resultant=data.axes[0].shape)
    offsets = data[f"{channel_name}_1_moment_1"].points

    if spline:
        spline = wt.kit.Spline(data.axes[0].points, offsets, **spline_kwargs)
        return spline(tune_points).clip(data.axes[1].min(), data.axes[1].max())
    if np.allclose(data.axes[0].points, tune_points):
        return offsets.clip(data.axes[1].min(), data.axes[1].max())
    elif np.allclose(data.axes[0].points, tune_points[::-1]):
        return offsets.clip(data.axes[1].min(), data.axes[1].max())[::-1]
    else:
        raise ValueError("Data points and instrument points do not match, and splining disabled")


def intensity(
    *,
    data,
    channel,
    arrangement,
    tune,
    instrument=None,
    level=False,
    gtol=0.01,
    ltol=0.1,
    autosave=True,
    save_directory=None,
    **spline_kwargs,
):
    """Workup a generic intensity plot for a single dependent.

    Parameters
    ----------
    data : wt.data.Data
        should be in (setpoint, dependent)
    channel: wt.data.Channel or int or str
        channel to process
    arrangement: str
        name of the arrangement to modify in the instrument
    tune: str
        name of the tune to modify in the instrument
    instrument: attune.Instrument, optional
        instrument object to modify (Default None: make a new instrument)
    level: bool, optional
        toggle leveling data (Defalts to False)
    gtol: float, optional
        global tolerance for rejecting noise level relative to global maximum
    ltol: float, optional
        local tolerance for rejecting data relative to slice maximum
    autosave: bool, optional
        toggles saving of instrument file and images (Defaults to True)
    save_directory: Path-like
        where to save (Defaults to current working directory)
    **spline_kwargs: optional
        extra arguments to pass to spline creation (e.g. s=0, k=1 for linear interpolation)

    Returns
    -------
    attune.Instrument
        New instrument object.
    """
    metadata = {
        "channel": channel,
        "arrangement": arrangement,
        "tune": tune,
        "level": level,
        "gtol": gtol,
        "ltol": ltol,
        "spline_kwargs": spline_kwargs,
    }
    if not isinstance(channel, (int, str)):
        metadata["channel"] = channel.natural_name
    transition = Transition("intensity", instrument, metadata=metadata, data=data)
    data = data.copy()
    data.convert("nm")
    if instrument is not None:
        old_instrument = instrument.as_dict()
        setpoints = instrument[arrangement][tune].independent
    else:
        old_instrument = None
        setpoints = data.axes[0].points
    # TODO: units
    setpoints.sort()

    if isinstance(channel, (int, str)):
        channel = data.channels[wt.kit.get_index(data.channel_names, channel)]
        orig_channel = data.create_channel(
            f"{channel.natural_name}_orig", channel, units=channel.units
        )

    # TODO: check if level does what we want
    if level:
        data.level(channel.natural_name, 0, -3)

    # TODO: gtol/ltol should maybe be moved to wt
    cutoff = channel.max() * gtol
    channel.clip(min=cutoff)
    max_axis = tuple(i for i, v in enumerate(data.axes[0].shape) if v > 1)
    cutoff = np.nanmax(channel[:], axis=1, keepdims=True) * ltol
    channel.clip(min=cutoff)

    offsets = _intensity(data, channel.natural_name, setpoints, **spline_kwargs)
    print(setpoints)
    print(offsets)
    try:
        raw_offsets = _intensity(data, channel.natural_name, setpoints, spline=False)
    except ValueError:
        raw_offsets = None

    units = data.axes[1].units
    if units == "None":
        units = None

    if instrument is not None:
        # Note, assumes units can be converted here... which I hope so
        if units is not None:
            offsets = wt.units.convert(
                offsets,
                units,
                old_instrument["arrangements"][arrangement]["tunes"][tune]["dep_units"],
            )
        old_instrument["arrangements"][arrangement]["tunes"][tune]["independent"] = setpoints
        old_instrument["arrangements"][arrangement]["tunes"][tune]["dependent"] += offsets
        try:
            del old_instrument["transition"]
        except KeyError:
            pass
        new_instrument = Instrument(**old_instrument, transition=transition)
    else:
        arr = Arrangement(arrangement, {tune: Tune(setpoints, offsets, dep_units=units)})
        new_instrument = Instrument(
            {arrangement: arr}, {tune: Setable(tune)}, transition=transition
        )

    fig, _ = plot_intensity(
        data, channel.natural_name, arrangement, tune, new_instrument, instrument, raw_offsets
    )

    if autosave:
        save(new_instrument, fig, "intensity", save_directory)
    return new_instrument



"""Workup an instrument such that the output is the expected color"""

import numpy as np
import WrightTools as wt

from ._instrument import Instrument
from ._arrangement import Arrangement
from ._setable import Setable
from ._tune import Tune
from ._transition import Transition
from ._plot import plot_setpoint
from ._common import save


# --- processing methods --------------------------------------------------------------------------

__all__ = ["setpoint"]


def _setpoint(data, channel_name, tune_points, *, spline=True, **spline_kwargs):
    offsets = []
    chops = data.chop(1)
    for c in chops.values():
        xi = c.axes[0].points
        yi = c[channel_name].points
        xi, yi = wt.kit.remove_nans_1D(xi, yi)
        if np.all(np.isnan(yi)):
            offsets.append(np.nan)
        elif np.nanmin(yi) <= 0 <= np.nanmax(yi):
            p = np.polynomial.Polynomial.fit(yi, xi, 2)
            offsets.append(p(0))
        else:
            offsets.append(np.nan)

    offsets = np.array(offsets)
    if spline:
        spline = wt.kit.Spline(data.axes[0].points, offsets, **spline_kwargs)
        return spline(tune_points).clip(data.axes[1].min(), data.axes[1].max())
    if np.allclose(data.axes[0].points, tune_points):
        return offsets[::-1].clip(data.axes[1].min(), data.axes[1].max())
    elif np.allclose(data.axes[0].points, tune_points[::-1]):
        return offsets[::-1].clip(data.axes[1].min(), data.axes[1].max())[::-1]
    else:
        raise ValueError("Data points and instrument points do not match, and splining disabled")


def setpoint(
    *,
    data,
    channel,
    arrangement,
    tune,
    instrument=None,
    autosave=True,
    save_directory=None,
    **spline_kwargs
):
    """Workup a generic setpoint plot for a single tune.

    Parameters
    ----------
    data : wt.data.Data object
        should be in (setpoint, tune)
    channel: wt.data.Channel or int or str
        channel to process
    arrangement: str
        name of the arrangement to modify
    tune: str
        name of the tune to modify in the instrument
    instrument: attune.Curve, optional
        instrument object to modify (Default None: make a new instrument)
    autosave: bool, optional
        toggles saving of instrument file and images (Defaults to True)
    save_directory: Path-like
        where to save (Defaults to current working directory)
    **spline_kwargs: optional
        extra arguments to pass to spline creation (e.g. s=0, k=1 for linear interpolation)

    Returns
    -------
    attune.Curve
        New instrument object.
    """
    metadata = {
        "channel": channel,
        "arrangement": arrangement,
        "tune": tune,
        "spline_kwargs": spline_kwargs,
    }
    if not isinstance(channel, (int, str)):
        metadata["channel"] = channel.natural_name
    transition = Transition("setpoint", instrument, metadata=metadata, data=data)
    data = data.copy()
    data.convert("nm")
    if instrument is not None:
        old_instrument = instrument.as_dict()
        setpoints = instrument[arrangement][tune].independent
    else:
        setpoints = data.axes[0].points
    # TODO: units
    setpoints.sort()

    if isinstance(channel, (int, str)):
        channel = data.channels[wt.kit.get_index(data.channel_names, channel)]

    dims = [1] * data.ndim
    dims[0] = setpoints.size  # TODO: be more robust, don't assume 0 index
    channel -= setpoints.reshape(dims)

    offsets = _setpoint(data, channel.natural_name, setpoints, **spline_kwargs)
    try:
        raw_offsets = _setpoint(data, channel.natural_name, setpoints, spline=False)
    except ValueError:
        raw_offsets = None

    if instrument is not None:
        old_instrument["arrangements"][arrangement]["tunes"][tune]["independent"] = setpoints
        old_instrument["arrangements"][arrangement]["tunes"][tune]["dependent"] += offsets
        try:
            del old_instrument["transition"]
        except KeyError:
            pass
        new_instrument = Instrument(**old_instrument, transition=transition)
    else:
        arr = Arrangement(arrangement, {tune: Tune(setpoints, offsets)})
        new_instrument = Instrument(
            {arrangement: arr}, {tune: Setable(tune)}, transition=transition
        )

    fig, _ = plot_setpoint(
        data, channel.natural_name, arrangement, tune, new_instrument, instrument, raw_offsets
    )

    if autosave:
        save(new_instrument, fig, "setpoint", save_directory)
    return new_instrument



__all__ = ["offset_by", "offset_to"]

import copy

from ._transition import Transition
from ._tune import Tune


def offset_by(instrument, arrangement, tune, amount, amount_units=None):
    md = {
        "arrangement": arrangement,
        "tune": tune,
        "amount": amount,
        "amount_units": amount_units,
    }
    to_offset = instrument[arrangement][tune]
    if amount_units is not None:
        amount = wt.units.convert(amount, amount_units, to_offset.dep_units)
    instr = copy.deepcopy(instrument)
    instr[arrangement]._tunes[tune] = Tune(
        to_offset.independent,
        to_offset.dependent + amount,
        dep_units=to_offset.dep_units,
    )
    instr._transition = Transition("offset_by", instrument, metadata=md)
    instr._load = None
    return instr


def offset_to(
    instrument,
    arrangement,
    tune,
    destination,
    setpoint,
    destination_units=None,
    setpoint_units=None,
):
    to_offset = instrument[arrangement][tune]
    current = to_offset(setpoint, ind_units=setpoint_units, dep_units=destination_units)
    offset = destination - current
    instr = offset_by(instrument, arrangement, tune, offset)
    md = {
        "arrangement": arrangement,
        "tune": tune,
        "destination": destination,
        "setpoint": setpoint,
        "destination_units": destination_units,
        "setpoint_units": setpoint_units,
    }
    instr._transition = Transition("offset_to", instrument, metadata=md)
    return instr



__all__ = ["Note"]


from typing import Dict, Union
from ._setable import Setable


class Note:
    def __init__(
        self,
        setables: Dict[str, Setable],
        setable_positions: Dict[str, Union[float, str]],
        arrangement_name: str,
    ):
        """A particular set of motor positions.

        Parameters
        ----------
        setables: Dict[str, Setable]
            The setables represented in the note
        setable_positions: Dict[str, Union[float, str]]
            Mapping of setable keys to positions
        arrangement_name: str
            The name of the arrangement used to make this note
        """
        self.setables: Dict[str, Setable] = setables
        self.setable_positions: Dict[str, Union[str, float]] = setable_positions
        self.arrangement_name: str = arrangement_name

    def __getitem__(self, k):
        return self.setable_positions[k]

    def __repr__(self):
        return f"Note({self.setables}, {self.setable_positions}, {repr(self.arrangement_name)})"

    def items(self):
        """Items in the Note."""
        return self.setable_positions.items()

    def keys(self):
        """Settable keys in the Note."""
        return self.setable_positions.keys()

    def values(self):
        """Settable values."""
        return self.setable_positions.values()



"""Internal plotting functions for workup."""

import WrightTools as wt
import matplotlib.pyplot as plt
import numpy as np


def plot_intensity(
    data, channel, arrangement, tune, instrument, prior_instrument=None, raw_offsets=None
):
    fig, gs = wt.artists.create_figure(
        width="single", nrows=2, cols=[1, "cbar"], default_aspect=0.5
    )
    ax = plt.subplot(gs[0, 0])
    instrument_plot_kwargs = {"lw": 5, "c": "k", "alpha": 0.5}
    prior_instrument_plot_kwargs = {"lw": 2, "c": "k"}
    new_tune = instrument[arrangement][tune]
    ax.plot(new_tune.independent, new_tune.dependent, **instrument_plot_kwargs)
    if prior_instrument:
        prior_tune = prior_instrument[arrangement][tune]
        ax.plot(
            prior_tune.independent,
            prior_tune.dependent,
            **prior_instrument_plot_kwargs,
        )
    wt.artists.plot_gridlines()
    ax.set_ylabel(tune)
    ax.set_xlim(instrument[arrangement].ind_min, instrument[arrangement].ind_max)
    ax.xaxis.set_tick_params(label1On=False)

    ax = plt.subplot(gs[1, 0])
    graymap = "greyscale"
    ax.pcolor(data, channel=f"{channel}_orig", cmap=graymap)
    ax.pcolor(data, channel=channel)
    if prior_instrument:
        ypoints = new_tune.dependent - prior_instrument(new_tune.independent, arrangement)[tune]
    else:
        ypoints = new_tune.dependent

    if raw_offsets is not None:
        ax.plot(new_tune.independent, raw_offsets, c="grey", lw=5, alpha=0.5)

    ax.plot(new_tune.independent, ypoints, **instrument_plot_kwargs)
    ax.axhline(0, **prior_instrument_plot_kwargs)
    wt.artists.plot_gridlines()
    ax.set_ylabel(rf"$\mathsf{{\Delta {tune}}}$")
    ax.set_xlabel(data.axes[0].label)
    ax.set_xlim(instrument[arrangement].ind_min, instrument[arrangement].ind_max)

    cax = plt.subplot(gs[1, 1])
    ticks = np.linspace(data[channel].null, data[channel].max(), 11)
    wt.artists.plot_colorbar(
        cax, vlim=(data[channel].null, data[channel].max()), ticks=ticks, label=channel
    )

    return fig, gs


def plot_setpoint(
    data, channel, arrangement, tune, instrument, prior_instrument=None, raw_offsets=None
):
    fig, gs = wt.artists.create_figure(
        width="single", nrows=2, cols=[1, "cbar"], default_aspect=0.5
    )
    ax = plt.subplot(gs[0, 0])
    instrument_plot_kwargs = {"lw": 5, "c": "k", "alpha": 0.5}
    prior_instrument_plot_kwargs = {"lw": 2, "c": "k"}
    ax.plot(
        instrument[arrangement][tune].independent,
        instrument[arrangement][tune].dependent,
        **instrument_plot_kwargs,
    )
    ax.set_xlim(instrument[arrangement].ind_min, instrument[arrangement].ind_max)
    ax.xaxis.set_tick_params(label1On=False)
    if prior_instrument:
        ax.plot(
            prior_instrument[arrangement][tune].independent,
            prior_instrument[arrangement][tune].dependent,
            **prior_instrument_plot_kwargs,
        )
    ax.set_ylabel(tune)
    wt.artists.plot_gridlines()

    ax = plt.subplot(gs[1, 0])
    # data[channel][:] -= data.axes[0][:]
    data[channel].signed = True
    limits = -0.05 * data[channel].mag(), 0.05 * data[channel].mag()
    ax.pcolor(data, channel=channel, vmin=limits[0], vmax=limits[1])
    ax.set_xlim(instrument[arrangement].ind_min, instrument[arrangement].ind_max)
    ax.set_ylim(data.axes[1].min(), data.axes[1].max())
    if prior_instrument:
        ypoints = (
            instrument[arrangement][tune].dependent
            - prior_instrument(instrument[arrangement][tune].independent, arrangement)[tune]
        )
    else:
        ypoints = instrument[arrangement][tune].dependent

    if raw_offsets is not None:
        ax.plot(
            instrument[arrangement][tune].independent, raw_offsets[::-1], c="grey", lw=5, alpha=0.5
        )
    ax.plot(instrument[arrangement][tune].independent, ypoints, **instrument_plot_kwargs)
    ax.axhline(0, **prior_instrument_plot_kwargs)
    wt.artists.plot_gridlines()
    ax.set_ylabel(rf"$\mathsf{{\Delta {tune}}}$")
    ax.set_xlabel(data.axes[0].label)
    ax.set_xlim(instrument[arrangement].ind_min, instrument[arrangement].ind_max)

    cax = plt.subplot(gs[1, 1])
    ticks = np.linspace(*limits, 11)
    wt.artists.plot_colorbar(cax, vlim=limits, ticks=ticks, label=channel, cmap="signed")

    return fig, gs


def plot_tune_test(data, channel, used_offsets, raw_offsets=None):
    fig, gs = wt.artists.create_figure(default_aspect=0.5, cols=[1, "cbar"])
    # heatmap
    ax = plt.subplot(gs[0, 0])
    ax.pcolor(data, channel=channel)
    ax.set_xlim(data.axes[0].min(), data.axes[0].max())
    # lines
    if raw_offsets is not None:
        ax.plot(data.axes[0].points, raw_offsets, c="grey", lw=5, alpha=0.5)

    ax.plot(
        data.axes[0].points,
        used_offsets,
        c="k",
        lw=5,
        alpha=0.5,
    )
    ax.axhline(c="k", lw=1)
    ax.grid()

    ax.set_xlabel(data.axes[0].label)
    ax.set_ylabel(rf"$\mathsf{{\Delta}}$ {data.axes[0].label}")

    # colorbar
    cax = plt.subplot(gs[:, -1])
    label = channel
    ticks = np.linspace(0, data[channel].max(), 7)
    wt.artists.plot_colorbar(cax=cax, label=label, ticks=ticks)

    return fig, gs


def plot_holistic(
    data,
    amp_channel,
    center_channel,
    arrangement,
    tunes,
    instrument,
    prior_instrument,
    raw_offsets=None,
):
    data[amp_channel].normalize()

    amp_cmap = wt.artists.colormaps["default"]
    center_cmap = wt.artists.colormaps["rainbow"]

    fig, gs = wt.artists.create_figure(nrows=2, cols=[1, "cbar"])
    ax_amp = plt.subplot(gs[0, 0])
    ax_cen = plt.subplot(gs[1, 0])

    cax_amp = plt.subplot(gs[0, 1])
    cax_center = plt.subplot(gs[1, 1])
    amp_ticks = np.linspace(0, 1, 11)
    center_ticks = instrument[arrangement].independent

    ax_amp.pcolor(data, channel=amp_channel, cmap=amp_cmap)
    ax_amp.contour(
        data,
        channel=center_channel,
        levels=sorted(center_ticks),
        cmap=center_cmap,
        linewidths=2,
        alpha=1,
        vmin=np.min(center_ticks),
        vmax=np.max(center_ticks),
    )
    ax_cen.pcolor(
        data,
        channel=center_channel,
        cmap=center_cmap,
        vmin=np.min(center_ticks),
        vmax=np.max(center_ticks),
    )
    ax_cen.contour(
        data,
        channel=amp_channel,
        levels=amp_ticks,
        cmap=amp_cmap,
        linewidths=2,
        alpha=1,
    )

    wt.artists.set_fig_labels(xlabel=data.axes[0].label, ylabel=data.axes[1].label)

    wt.artists.plot_colorbar(cax_amp, cmap=amp_cmap, ticks=amp_ticks, label="Intensity")
    wt.artists.plot_colorbar(cax_center, cmap=center_cmap, ticks=center_ticks, label="Center")
    cax_center.set_xlabel("")

    for i in range(2):
        axis = np.array(np.broadcast_to(data.axes[i][:], data[amp_channel].shape))
        axis[np.isnan(data[amp_channel])] = np.nan

        amin = min(
            np.min(instrument[arrangement][tunes[i]].dependent),
            np.min(prior_instrument[arrangement][tunes[i]].dependent),
            np.nanmin(axis),
        )
        amax = max(
            np.max(instrument[arrangement][tunes[i]].dependent),
            np.max(prior_instrument[arrangement][tunes[i]].dependent),
            np.nanmax(axis),
        )
        arange = amax - amin
        amin -= arange * 0.05
        amax += arange * 0.05
        if i == 0:
            ax_amp.set_xlim((amin, amax))
            ax_cen.set_xlim((amin, amax))
        else:
            ax_amp.set_ylim((amin, amax))
            ax_cen.set_ylim((amin, amax))

    for ax in [ax_amp, ax_cen]:
        ax.plot(
            prior_instrument[arrangement][tunes[0]].dependent,
            prior_instrument[arrangement][tunes[1]].dependent,
            color="k",
            linewidth=2,
            zorder=2,
        )
        ax.plot(
            instrument[arrangement][tunes[0]].dependent,
            instrument[arrangement][tunes[1]].dependent,
            color="k",
            linewidth=6,
            alpha=0.5,
            zorder=2,
        )
        if raw_offsets is not None:
            ax.scatter(
                raw_offsets[:, 0],
                raw_offsets[:, 1],
                color="w",
                s=50,
                zorder=10,
                marker="*",
            )  # TODO don't be bad about point handleing

    return fig, gs



from ._instrument import Instrument
from ._transition import Transition


def update_merge(base: Instrument, replace: Instrument) -> Instrument:
    """Update the base instrument with tunes from the replace instrument

    Name and any tunes not present in replace remain from base.
    Base is the instrument which is identified by the transition.
    If new arrangements are present in replace, they are added to base.
    """
    instr = base.as_dict()
    for arr_name, arr in replace.arrangements.items():
        if arr_name not in base.arrangements:
            instr["arrangements"][arr_name] = arr
            continue
        for tune_name, tune in arr.items():
            instr["arrangements"][arr_name]["tunes"][tune_name] = tune
    transition = Transition(type="update_merge", previous=base)
    instr.pop("transition", None)
    print(repr(transition))
    return Instrument(**instr, transition=transition)



"""Methods for processing OPA 800 tuning data."""


import numpy as np

import WrightTools as wt

from ._discrete_tune import DiscreteTune
from ._instrument import Instrument
from ._transition import Transition
from ._plot import plot_tune_test
from ._common import save
from ._map import map_ind_points

__all__ = ["tune_test"]


def _offsets(data, channel_name, tune_points, *, spline=True, **spline_kwargs):
    data.moment(axis=1, channel=channel_name, moment=1, resultant=data.axes[0].shape)
    offsets = data[f"{channel_name}_1_moment_1"].points

    if spline:
        return wt.kit.Spline(data.axes[0].points, offsets, **spline_kwargs)
    if np.allclose(data.axes[0].points, tune_points):
        return offsets.clip(data.axes[1].min(), data.axes[1].max())
    if np.allclose(data.axes[0].points, tune_points[::-1]):
        return offsets.clip(data.axes[1].min(), data.axes[1].max())[::-1]
    else:
        raise ValueError("Data points and instrument points do not match, and splining disabled")


def tune_test(
    *,
    data,
    channel,
    arrangement,
    instrument,
    level=False,
    gtol=0.01,
    ltol=0.1,
    restore_setpoints=True,
    autosave=True,
    save_directory=None,
    **spline_kwargs,
) -> Instrument:
    """Workup a Tune Test.

    Parameters
    ----------
    data : wt.data.Data
        should be in (setpoint, dependent)
    channel: wt.data.Channel or int or str
        channel to process
    arrangement: str
        name of the arrangment to modify
    instrument: attune.Instrument
        instrument object to modify
    level: bool, optional
        toggle leveling data (Defalts to False)
    gtol: float, optional
        global tolerance for rejecting noise level relative to global maximum
    ltol: float, optional
        local tolerance for rejecting data relative to slice maximum
    restore_setpoints: bool, optional
        toggles remapping onto original setpoints for each tune (default is True)
    autosave: bool, optional
        toggles saving of instrument file and images (Defaults to True)
    save_directory: Path-like
        where to save (Defaults to current working directory)
    **spline_kwargs: optional
        extra arguments to pass to spline creation (e.g. s=0, k=1 for linear interpolation)

    Returns
    -------
    attune.Instrument
        New instrument object.

    Example
    -------
    ```
    >>> d = wt.open(path_to_data_file)
    >>> d.transform("w3_points", "wm-w3")
    >>> instr2 = attune.tune_test(
            data=d,
            channel="sample",
            arrangement="df1",
            instrument=instr,
            gtol=.0002,
            ltol=.01,
            autosave=False,
            save_directory=folder
        )
    ```

    """
    metadata = {
        "channel": channel,
        "arrangement": arrangement,
        "level": level,
        "gtol": gtol,
        "ltol": ltol,
        "spline_kwargs": spline_kwargs,
    }
    if not isinstance(channel, (int, str)):
        metadata["channel"] = channel.natural_name
    transition = Transition("tune_test", instrument, metadata=metadata, data=data)

    data = data.copy()
    data.convert("nm")

    setpoints = data.axes[0].points
    setpoints.sort()

    if isinstance(channel, (int, str)):
        channel = data.channels[wt.kit.get_index(data.channel_names, channel)]
        orig_channel = data.create_channel(
            f"{channel.natural_name}_orig", channel, units=channel.units
        )

    # TODO: check if level does what we want
    if level:
        data.level(channel.natural_name, 0, -3)

    # TODO: gtol/ltol should maybe be moved to wt
    cutoff = channel.max() * gtol
    channel.clip(min=cutoff)
    max_axis = tuple(i for i, v in enumerate(data.axes[0].shape) if v > 1)
    cutoff = np.nanmax(channel[:], axis=1, keepdims=True) * ltol
    channel.clip(min=cutoff)

    offset_spline = _offsets(data, channel.natural_name, setpoints, **spline_kwargs)
    try:
        raw_offsets = _offsets(data, channel.natural_name, setpoints, spline=False)
    except ValueError:
        raw_offsets = None

    old_instrument = instrument.as_dict()
    for tune in old_instrument["arrangements"][arrangement]["tunes"].values():
        if "ranges" in tune:
            # Discrete tune in dict form
            continue
        tune["independent"] += offset_spline(tune["independent"])
    new_instrument = Instrument(**old_instrument)

    if restore_setpoints:
        for tune in new_instrument[arrangement].keys():
            if isinstance(instrument[arrangement][tune], DiscreteTune):
                continue
            new_instrument = map_ind_points(
                new_instrument, arrangement, tune, instrument[arrangement][tune].independent
            )

    new_instrument._transition = transition

    fig, _ = plot_tune_test(
        data,
        channel.natural_name,
        used_offsets=offset_spline(setpoints),
        raw_offsets=raw_offsets,
    )

    if autosave:
        save(new_instrument, fig, "tune_test", save_directory)
    return new_instrument



__all__ = ["Tune"]


import WrightTools as wt
import numpy as np
import scipy.interpolate


class Tune:
    def __init__(self, independent, dependent, *, dep_units=None, **kwargs):
        """A Tune which maps one set of inputs to associated output points.

        Currently all tunes are assumed to have "nm" as their independent array units.
        All mappings are linear interpolations

        Parameters
        ----------
        independent: 1D array-like
            The independent axis for input values to be mapped.
            Must be the same shape as dependent.
        dependent: 1D array-like
            The depending axis for the mapping.
            Must be the same shape as independent.
        dep_units: str (optional)
            Units for the dependent axis

        Note: kwargs are provided to make the serialized dictionary with ind_units
        easy to initialize into a Tune object, but are currently ignored.
        """
        independent = np.asarray(independent)
        dependent = np.asarray(dependent)
        assert independent.size == dependent.size
        assert independent.ndim == dependent.ndim == 1
        self._ind_units = "nm"
        self._dep_units = dep_units
        self._interp = scipy.interpolate.interp1d(independent, dependent, fill_value="extrapolate")

    def __repr__(self):
        if self.dep_units is None:
            return f"Tune({repr(self.independent)}, {repr(self.dependent)})"
        return f"Tune({repr(self.independent)}, {repr(self.dependent)}, dep_units={repr(self.dep_units)})"

    def __call__(self, ind_value, *, ind_units=None, dep_units=None):
        if ind_units is not None and self._ind_units is not None:
            ind_value = wt.units.convert(ind_value, ind_units, self._ind_units)
        ret = self._interp(ind_value)
        if dep_units is not None and self._dep_units is not None:
            ret = wt.units.convert(ret, self._dep_units, dep_units)
        return ret

    def __len__(self):
        return len(self.independent)

    def __eq__(self, other):
        if not np.allclose(self.independent, other.independent):
            return False
        if not np.allclose(self(self.independent), other(other.independent)):
            return False
        return self.ind_units == other.ind_units and self.dep_units == other.dep_units

    def as_dict(self):
        """Serialize this Tune as a python dictionary."""
        out = {}
        out["independent"] = list(self.independent)
        out["dependent"] = list(self.dependent)
        out["ind_units"] = self.ind_units
        out["dep_units"] = self.dep_units
        return out

    @property
    def independent(self):
        """The independent (input) values for the tune points."""
        return self._interp.x.astype(float)

    @property
    def dependent(self):
        """The dependent (output) values for the tune points."""
        return self._interp.y.astype(float)

    @property
    def ind_max(self):
        """The maximum independent (input) value for the tune."""
        return self.independent.max()

    @property
    def ind_min(self):
        """The minimum independent (input) value for the tune."""
        return self.independent.min()

    @property
    def ind_units(self):
        """The units of the independent (input) values."""
        return self._ind_units

    @property
    def dep_units(self):
        """The units of the dependent (output) values."""
        return self._dep_units



__all__ = ["Arrangement"]


from typing import Dict, Union

import numpy as np

from ._tune import Tune
from ._discrete_tune import DiscreteTune


def mktune(dict_):
    if "ranges" in dict_:
        return DiscreteTune(**dict_)
    return Tune(**dict_)


class Arrangement:
    def __init__(self, name: str, tunes: Dict[str, Union[DiscreteTune, Tune, dict]]):
        """Arrangement of several Tunes to form one cohesive set.

        Tunes may represent either motors or other arrangements, however
        semantic meaning is provided by the Instrument which contains the
        Arrangement, to the arrangement, they are all string keys mapped to tunes.

        All tunes *must* have the same independent units, and must overlap.

        Parameters
        ----------
        name: str
            A name for this Arrangement, used to identify this Arrangement
            from other Arrangements which may depend on this one.
        tunes: Dict[str, Tune]
            Mapping of names to Tune objects which compose the Arrangement
        """
        self._name: str = name
        self._tunes: Dict[str, Union[DiscreteTune, Tune]] = {
            k: mktune(v) if isinstance(v, dict) else v for k, v in tunes.items()
        }
        self._ind_units: str = "nm"

    def __repr__(self):
        return f"Arrangement({repr(self.name)}, {repr(self.tunes)})"

    def __getitem__(self, key):
        return self.tunes[key]

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.tunes != other.tunes:
            return False
        return True

    @property
    def independent(self):
        """Returns a 1-dimensional numpy array with the set of all unique independent points.

        Points closer together than 1/1000th of the total dynamic range are considered identical.

        Only returns points within range of all tunes.
        """
        out = np.unique(
            np.concatenate([t.independent for t in self._tunes.values() if isinstance(t, Tune)], 0)
        )
        tol = 1e-3 * (self.ind_max - self.ind_min)
        diff = np.append(tol * 2, np.diff(out))
        out = out[diff > tol]
        out = out[out <= self.ind_max]
        out = out[out >= self.ind_min]
        return out

    def keys(self):
        """Return the names of the tunes in the arrangment."""
        return self.tunes.keys()

    def values(self):
        """Return the tunes in the arrangment."""
        return self.tunes.values()

    def items(self):
        """Return the names and tunes in the arrangment."""
        return self.tunes.items()

    def as_dict(self):
        """Dictionary representation of the Arrangement"""
        out = {}
        out["name"] = self._name
        out["tunes"] = {k: v.as_dict() for k, v in self._tunes.items()}
        return out

    @property
    def ind_max(self):
        """The maximum independant (input) value for this arrangement."""
        return min([t.ind_max for t in self._tunes.values() if isinstance(t, Tune)])

    @property
    def ind_min(self):
        """The minimum independant (input) value for this arrangement."""
        return max([t.ind_min for t in self._tunes.values() if isinstance(t, Tune)])

    @property
    def name(self):
        """The name of the arrangement."""
        return self._name

    @property
    def tunes(self):
        """The tunes in the arrangement."""
        return self._tunes



__all__ = ["Setable"]

from typing import Union, Optional
from dataclasses import dataclass, asdict


@dataclass
class Setable:
    """Container for default positions for a independent variable (something "setable")

    Parameters
    ----------
    name: str
        The key for a independent variable
    default: str or float (optional)
        The default value for this independent variable.  Defaults to None (no default).
    """

    name: str
    default: Optional[Union[str, float]] = None

    def as_dict(self):
        return asdict(self)



"""Function for processing multi-dependent tuning data."""

import itertools

import numpy as np
import scipy

import WrightTools as wt
from ._instrument import Instrument
from ._transition import Transition
from ._plot import plot_holistic
from ._common import save


__all__ = ["holistic"]


def _holistic(data, amplitudes, centers, arrangement):
    points = np.array([np.broadcast_to(a[:], amplitudes.shape).flatten() for a in data.axes]).T
    ndim = len(data.axes)
    delaunay = scipy.spatial.Delaunay(points)

    amp_interp = scipy.interpolate.LinearNDInterpolator(delaunay, amplitudes.points.flatten())
    cen_interp = scipy.interpolate.LinearNDInterpolator(delaunay, centers.points.flatten())

    # def
    out_points = []
    for p in arrangement.independent:
        iso_points = []
        for s, pts, vals in _find_simplices_containing(delaunay, cen_interp, p):
            iso_points.extend(_edge_intersections(pts, vals, p))
        iso_points = np.array(iso_points)
        if len(iso_points) > 3:
            out_points.append(
                tuple(_fit_gauss(iso_points.T[i], amp_interp(iso_points)) for i in range(ndim))
            )
        else:
            out_points.append(tuple(np.nan for i in range(ndim)))

    return np.array(out_points)


def holistic(
    *,
    data,
    channels,
    arrangement,
    tunes,
    instrument,
    spectral_axis=-1,
    level=False,
    gtol=0.01,
    autosave=True,
    save_directory=None,
    **spline_kwargs,
):
    """Workup multi-dependent tuning data.

    Note:
    At this time, this function expects 2-dimensional motor space.
    The algorithm should generalize to N-dimensional motor space,
    however this is untested and plotting likely will fail.

    Parameters
    ----------
    data: WrightTools.Data
        The data object to process.
    channels: WrightTools.data.Channel or int or str or 2-tuple
        If singular: the spectral axis, from which the 0th and 1st moments will be taken to
        obtain amplitudes and centers. In this case, `spectral_axis` determines which axis is
        used to obtain the moments.
        If a tuple: (amplitudes, centers), then these channels will be used directly.
    tunes: iterable of str
        Names of the tunes to modify in the instrument, in the same order as the axes of `data`.
        Must not be DiscreteTunes.
    instrument: attune.Instrument
        Instrument object to modify. Setpoints are determined from the instrument.

    Keyword Parameters
    ------------------
    spectral_axis: WrightTools.data.Axis or int or str (default -1)
        The axis along which to take moments.
        Only applies if a single channel is given.
    level: bool (default False)
        Toggle leveling data. If two channels are given, only the amplitudes are leveled.
        If a single channel is given, leveling occurs before taking the moments.
    gtol: float (default 0.01)
        Global tolerance for rejecting noise level relative to the global maximum.
    autosave: bool (default True)
        Toggles saving of instrument files and images.
    save_directory: Path-like (Defaults to current working directory)
        Specify where to save files.
    **spline_kwargs:
        Extra arguments to pass to spline creation (e.g. s=0, k=1 for linear interpolation)
    """
    metadata = {
        "channels": channels,
        "arrangement": arrangement,
        "tunes": tunes,
        "spectral_axis": spectral_axis,
        "level": level,
        "gtol": gtol,
        "spline_kwargs": spline_kwargs,
    }

    if not isinstance(channels, (int, str)):
        try:
            metadata["channels"] = list(channels)
            if not isinstance(channels[0], (int, str)):
                metadata["channels"][0] = channels[0].natural_name
            if not isinstance(channels[1], (int, str)):
                metadata["channels"][1] = channels[1].natural_name
        except TypeError:
            metadata["channels"] = channel.natural_name
    transition = Transition("holistic", instrument, metadata=metadata, data=data)

    # collect
    data = data.copy()

    if isinstance(channels, (str, wt.data.Channel)):
        if level:
            data.level(channels, 0, -3)
        if isinstance(spectral_axis, int):
            spectral_axis = data.axis_names[spectral_axis]
        elif isinstance(spectral_axis, wt.data.Axis):
            spectral_axis = spectral_axis.expression
        getattr(data, spectral_axis).convert("nm")
        # take channel moments
        data.moment(
            axis=spectral_axis,
            channel=channels,
            resultant=wt.kit.joint_shape(*[a for a in data.axes if a.expression != spectral_axis]),
            moment=0,
        )
        data.moment(
            axis=spectral_axis,
            channel=channels,
            resultant=wt.kit.joint_shape(*[a for a in data.axes if a.expression != spectral_axis]),
            moment=1,
        )
        amplitudes = data.channels[-2]
        centers = data.channels[-1]
        data.transform(*[a for a in data.axis_expressions if a != spectral_axis])
    else:
        amplitudes, centers = channels
        if isinstance(amplitudes, (int, str)):
            amplitudes = data.channels[wt.kit.get_index(data.channel_names, amplitudes)]
        if isinstance(centers, (int, str)):
            centers = data.channels[wt.kit.get_index(data.channel_names, centers)]
        if level:
            data.level(amplitudes.natural_name, 0, -3)

    if gtol is not None:
        cutoff = amplitudes.max() * gtol
        amplitudes.clip(min=cutoff)
    centers[np.isnan(amplitudes)] = np.nan

    out_points = _holistic(data, amplitudes, centers, instrument[arrangement])
    splines = [
        wt.kit.Spline(instrument[arrangement].independent, vals, **spline_kwargs)
        for vals in out_points.T
    ]

    new_instrument = _gen_instr(instrument, arrangement, tunes, splines, transition)

    fig, _ = plot_holistic(
        data,
        amplitudes.natural_name,
        centers.natural_name,
        arrangement,
        tunes,
        new_instrument,
        instrument,
        out_points,
    )

    if autosave:
        save(new_instrument, fig, "holistic", save_directory)
    return new_instrument


def _gen_instr(instrument, arrangement, tunes, splines, transition):
    new_instrument = instrument.as_dict()
    del new_instrument["transition"]
    setpoints = instrument[arrangement].independent
    for tune, spline in zip(tunes, splines):
        new_instrument["arrangements"][arrangement]["tunes"][tune]["independent"] = setpoints
        new_instrument["arrangements"][arrangement]["tunes"][tune]["dependent"] = spline(setpoints)
    return Instrument(**new_instrument, transition=transition)


def _find_simplices_containing(delaunay, interpolator, point):
    for s in delaunay.simplices:
        extrema = interpolator([p for p in delaunay.points[s]])
        if min(extrema) < point <= max(extrema):
            yield s, delaunay.points[s], extrema


def _edge_intersections(points, evaluated, target):
    sortord = np.argsort(evaluated)
    evaluated = evaluated[sortord]
    points = points[sortord]
    for (p1, p2), (v1, v2) in zip(
        itertools.combinations(points, 2), itertools.combinations(evaluated, 2)
    ):
        if v1 < target <= v2:
            yield tuple(
                p1[i] + (p2[i] - p1[i]) * ((target - v1) / (v2 - v1)) for i in range(len(p1))
            )


def _fit_gauss(x, y):
    x, y = wt.kit.remove_nans_1D(x, y)

    def resid(inps):
        nonlocal x, y
        return y - _gauss(*inps)(x)

    bounds = [(-np.inf, np.inf) for i in range(3)]
    x_range = np.max(x) - np.min(x)
    bounds[0] = (np.min(x) - x_range / 10, np.max(x) + x_range / 10)
    bounds = np.array(bounds).T
    x0 = [np.median(x), x_range / 10, np.max(y)]
    opt = scipy.optimize.least_squares(resid, x0, bounds=bounds)
    return opt.x[0]


def _gauss(center, sigma, amplitude):
    return lambda x: amplitude * np.exp(-1 / 2 * (x - center) ** 2 / sigma**2)



__all__ = ["DiscreteTune"]

from typing import Dict, Tuple, Optional

import WrightTools as wt
import numpy as np


class DiscreteTune:
    def __init__(
        self, ranges: Dict[str, Tuple[float, float]], default: Optional[str] = None, **kwargs
    ):
        """A Tune which maps one set of inputs to associated output points.

        Currently all tunes are assumed to have "nm" as their independent units.

        Parameters
        ----------
        ranges: dict[str, tuple[float, float]]
            dictionary mapping the key (string identifier of a discrete position)
            to a 2-tuple of (min, max) for the range for which that identifier should be used.
            This dict is ordered, the first result with a matching range (inclusive of boundaries)
            will be the one returned when called.
        default: Optional[str]
            The result to return if no matching range is represented.
            Default is None

        Note: kwargs are provided to make the serialized dictionary with ind_units
        easy to initialize into a DiscreteTune object, but are currently ignored.
        """
        self._ind_units = "nm"
        self._ranges = {k: tuple(v) for k, v in ranges.items()}
        self._default = default

    def __repr__(self):
        return f"DiscreteTune({repr(self.ranges)}, {repr(self.default)})"

    def __call__(self, ind_value, *, ind_units=None):
        """Evaluate the DiscreteTune at specific independent value(s).

        Paramters
        ---------
        ind_val: float-like or ndarray
            The value or values at which to evaluate the DiscreteTune.
        ind_units: Optional[str]
            Units of the independent variable.  Default is "nm".

        Returns
        -------
        key: str or ndarray
            The string identifier for the independent value.
            For an array of ind_val, an array of identifiers is given.

        """
        if ind_units is not None and self._ind_units is not None:
            ind_value = wt.units.convert(ind_value, ind_units, self._ind_units)
        if isinstance(ind_value, np.ndarray):
            out = np.full(
                ind_value.shape,
                self.default,
                dtype=f"U{max([len(s) for s in self.ranges.keys()])}",
            )
            for key, (imin, imax) in self.ranges.items():
                out[(ind_value >= imin) & (ind_value <= imax)] = key
            return out
        else:
            for key, (imin, imax) in self.ranges.items():
                if imin <= ind_value <= imax:
                    return key
            return self.default

    def __eq__(self, other):
        return self.ranges == other.ranges and self.default == other.default

    def as_dict(self):
        """Serialize this Tune as a python dictionary."""
        out = {}
        out["ranges"] = self.ranges
        out["ind_units"] = self.ind_units
        out["default"] = self.default
        return out

    @property
    def ranges(self):
        """The ranges for discrete setpoints."""
        return self._ranges

    @property
    def ind_units(self):
        """The units of the independent (input) values."""
        return self._ind_units

    @property
    def default(self):
        """The value returned if no supplied range applies."""
        return self._default



#! /usr/bin/env python3


# --- import -------------------------------------------------------------------------------------


import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(here, fname)).read()


here = os.path.abspath(os.path.dirname(__file__))

extra_files = {"attune": ["VERSION"]}

with open(os.path.join(here, "attune", "VERSION")) as version_file:
    version = version_file.read().strip()


# --- setup function -----------------------------------------------------------------------------

setup(
    name="attune",
    packages=find_packages(exclude=("tests", "tests.*")),
    package_data=extra_files,
    python_requires=">=3.7",
    install_requires=[
        "WrightTools>=3.2.5",
        "numpy",
        "scipy",
        "matplotlib",
        "appdirs",
        "python-dateutil",
        "maya",
    ],
    extras_require={
        "dev": ["black", "pre-commit", "pytest", "pytest-cov"],
        "docs": ["sphinx-gallery>0.3.0", "sphinx", "sphinx-rtd-theme"],
    },
    version=version,
    description="Tools for tuning optical parametric amplifiers and multidimensional spectrometers.",
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    author="Blaise Thompson",
    author_email="blaise@untzag.com",
    license="MIT",
    url="https://github.com/wright-group/attune",
    keywords="spectroscopy science multidimensional visualization",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
    ],
)



return__all__ = ["Transition", "TransitionType"]

from enum import Enum
from typing import Any, Optional, Dict, TYPE_CHECKING


if TYPE_CHECKING:
    import WrightTools as wt

    from ._instrument import Instrument


class TransitionType(str, Enum):
    create = "create"
    read = "read"
    restore = "restore"
    rename = "rename"
    offset_to = "offset_to"
    offset_by = "offset_by"
    map_limits = "map_limits"
    map_ind_points = "map_ind_points"
    tune_test = "tune_test"
    intensity = "intensity"
    setpoint = "setpoint"
    holistic = "holistic"
    update_merge = "update_merge"


class Transition:
    def __init__(
        self,
        type: TransitionType,
        previous: Optional["Instrument"] = None,
        metadata: Optional[Dict[str, Any]] = None,
        data: Optional["wt.Data"] = None,
    ):
        """Represent one processing step of an instrument.

        Parameters
        ----------
        type: TransitionType
            Indentity of the type of transition
        previous: Optional["Instrument"]
            The instrument which was modified in the transition.
        metadata: Optional[Dict[str, Any]]
            JSON serializable metadata associated with the transition.
        data: Optional["wt.Data"]
            A WrightTools Data object that was used to generate the transition.
        """
        self.type = type
        self.previous = previous
        if metadata is None:
            metadata = {}
        self.metadata = metadata
        self.data = data

    def __repr__(self):
        return f"Transition({repr(self.type)}, {repr(self.previous)}, {repr(self.metadata)})"

    def as_dict(self) -> Dict[str, Any]:
        """JSON serializable representation of the transition."""
        return {
            "type": self.type,
            "metadata": self.metadata,
        }



__all__ = ["from_topas4"]

import json
import os
import re

from .. import Arrangement
from .. import DiscreteTune
from .. import Instrument
from .. import Setable
from .. import Tune


def from_topas4(topas4_folder):
    """Convert a LightConversion marked up Topas4 calibration JSON file into an Instrument object.
    Topas4 Folder must contain at least the following 2 files:  OpticalDevices.json and Motors.json."""
    with open(os.path.join(topas4_folder, "OpticalDevices.json"), "r") as f:
        opt_dev = json.load(f)

    with open(os.path.join(topas4_folder, "Motors.json"), "r") as g:
        motors = json.load(g)["Motors"]

    with open(os.path.join(topas4_folder, "SeparationDevices.json"), "r") as g:
        sep_dev = json.load(g)

    opt_dev_active_guid = opt_dev.get("ActiveConfigurationGUID")
    opt_dev_conf = opt_dev["Configurations"]

    ind = 0
    for a in opt_dev_conf:
        if a["GUID"] == opt_dev_active_guid:
            opt_dev_active = a["OpticalDevices"]
            break
    else:
        raise ValueError("Active optical device GUID not found")

    sep_dev_active_guid = sep_dev.get("ActiveConfigurationGUID")
    sep_dev_conf = sep_dev["Configurations"]

    ind = 0
    for a in sep_dev_conf:
        if a["GUID"] == sep_dev_active_guid:
            sep_dev_active = a["SeparationDevices"]
            break
    else:
        raise ValueError("Active separation device GUID not found")

    arrangements = {}
    setables = {}

    for interaction in opt_dev_active:

        motorlist = {}
        for motor in motors:
            index = motor["Index"]
            motorlist[index] = motor["Title"]

        neutral_positions = interaction["NeutralPositions"]
        for pos in neutral_positions:
            index = pos["MotorIndex"]
            mot_name = motorlist[index]
            if pos["UseNamedPositionToResolveValue"]:
                position = pos["NamedPosition"]
            else:
                position = pos["Position"]
            setables[mot_name] = Setable(mot_name, position)

        interaction = interaction["Interactions"]

        for independent in interaction:
            arrange_name_full = independent.get("Type")

            if ">" in arrange_name_full:
                arrange_name, parent = arrange_name_full.split(">", maxsplit=1)
            else:
                arrange_name = arrange_name_full
                parent = arrange_name_full.split("-", maxsplit=1)[-1]
                if arrange_name == parent:
                    parent = None

            tunes = {}

            if parent is not None:
                inputpoints = independent.get("InputPoints")
                deparr = []
                indarr = []
                for inputpoint in inputpoints:
                    indarr.append(inputpoint.get("Output"))
                    deparr.append(inputpoint.get("Input"))

                tune = Tune(indarr, deparr)
                tunes[parent] = tune

            ind_motors = independent.get("MotorPositionCurves")
            for points in ind_motors:
                motorindex = points["MotorIndex"]
                k = motorlist[motorindex]
                deparr = []
                indarr = []
                for point in points["Points"]:
                    indarr.append(point.get("Input"))
                    deparr.append(point.get("Output"))

                tune = Tune(indarr, deparr)
                tunes[k] = tune

            arrangements[arrange_name] = Arrangement(arrange_name, tunes).as_dict()

    discrete_tunes = {x: {} for x in arrangements}
    for interaction in sep_dev_active:
        for mot_position in interaction["MotorPositionCurves"]:
            mot_index = mot_position["Index"]
            mot_name = motorlist[mot_index]
            for point in mot_position["Points"]:
                interaction = point["InteractionType"]
                interaction = interaction.replace("[", "")
                interaction = interaction.replace("]", "")
                interaction = interaction.replace("*", ".*")
                for arr, mots in discrete_tunes.items():
                    if re.match(interaction, arr):
                        x = mots.get(mot_name, {})
                        x[point["NamedPosition"]] = (point["Input"]["From"], point["Input"]["To"])
                        mots[mot_name] = x

    for arr, mots in discrete_tunes.items():
        for mot_name, pos in mots.items():
            tune = DiscreteTune(pos)
            arrangements[arr]["tunes"][mot_name] = tune

    instr = Instrument(arrangements, setables)

    return instr
